--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Purpose: Unmerge each input packet into output packets of length g_pkt_len.
-- Description:
-- . The merged packet length of the snk_in input packets must be an integer
--   multiple of g_pkt_len. The number of output packets g_nof_pkt_max is only
--   used to determine the maximum supported pkt_cnt range. The actual number
--   of output packets has to be <= g_nof_pkt_max, and is determined by input
--   packet length / g_pkt_len. Hence the dp_packet_unmerge can dynamically
--   handle different sizes of input packets, provided that their length is an
--   integer multiple of g_pkt_len.
-- . The pkt_cnt is passed on as src_out.channel index.
-- . The g_bsn_increment sets the BSN increment for the unmerged output
--   packets relative to the BSN of the snk_in input packet. When
--   g_bsn_increment = 0, then all unmerged output packets get the same BSN as
--   the input packet.
-- . The input snk_in.err and snk_in.empty are valid at the snk_in.eop, but
--   that is too late to apply them to the unmerged packets. Therefor assume
--   that the snk_in.err and snk_in_eop are already valid at the snk_in.sop
--   and remain valid until the snk_in.eop. Hence these signals are then valid
--   when snk_in.valid = '1'. Use same snk_in.err and snk_in.empty value for
--   all unmerged packets.
--                      _                       _                       _
--   snk_in.sop   _____| |_____________________| |_____________________| |_
--                                            _                       _
--   snk_in.eop   ___________________________| |_____________________| |___
--                      _       _       _       _       _       _       _
--   src_out.sop  _____|0|_____|1|_____|2|_____|0|_____|1|_____|2|_____|0|_
--                            _       _       _       _       _       _
--   src_out.eop  ___________|0|_____|1|_____|2|_____|0|_____|1|_____|2|___
--
-- . Flow control.
--   . The component does not add additional snk_out.ready flow control itself,
--     but it can listen to external src_in.ready flow control.
--   - If g_use_src_in_ready = false, then src_in is ignored, so no flow
--     control, else do use flow control by src_in.ready.
--   - If g_use_src_in_ready = true, then g_use_dp_latency_adapter = false
--     determines that the src_in.ready is used combinatorially else the
--     src_in.ready is used registered.
--   . The p_reg increases the ready latency (RL) by 1, because p_comb does
--     not use src_in.ready, instead p_comb drives d.src_out based on the
--     snk_in.sop, eop, valid. Therefore g_use_dp_latency_adapter = true is
--     needed to restore the RL from 2 to 1 for the src_out the output. The
--     alternative is to use a dp_pipeline to register d.src_out in
--     combination with src_in.ready at the input, to keep RL at 1.
--     The p_reg keeps the internal state of the component. With dp_pipeline
--     a separate output register is needed to store src_out. With the
--     dp_latency_adapter a feq register stages are needed to store src_out,
--     but the adantage of dp_latency_adapter is that it also registers the
--     src_in.ready so that may ease timing closure.

library IEEE,common_lib;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use common_lib.common_pkg.all;
use work.dp_stream_pkg.all;

entity dp_packet_unmerge is
  generic (
    g_use_src_in_ready         : boolean := false;
    g_use_dp_latency_adapter   : boolean := true;
    g_nof_pkt_max              : natural := 1;  -- Maximum nof packets to unmerge each incoming packet to
    g_pkt_len                  : natural := 1;  -- Length of the unmerged packets
    g_bsn_increment            : natural := 0
  );
  port (
    rst         : in  std_logic;
    clk         : in  std_logic;

    snk_out     : out t_dp_siso;
    snk_in      : in  t_dp_sosi;

    src_in      : in  t_dp_siso := c_dp_siso_rst;
    src_out     : out t_dp_sosi
  );
end dp_packet_unmerge;

architecture rtl of dp_packet_unmerge is
  type t_reg is record
    pkt_cnt     : natural range 0 to g_nof_pkt_max + 1;
    val_cnt     : natural range 0 to g_pkt_len + 1;
    src_out     : t_dp_sosi;
  end record;

  constant c_reg_rst : t_reg := (0, 0, c_dp_sosi_rst);

  signal r : t_reg;
  signal d : t_reg;

  -- Signals for g_use_dp_latency_adapter = true
  signal dp_latency_adapter_snk_out : t_dp_siso;
  signal dp_latency_adapter_snk_in  : t_dp_sosi;
  signal dp_latency_adapter_src_in  : t_dp_siso;
  signal dp_latency_adapter_src_out : t_dp_sosi;
begin
  -- Map t_reg outputs to entity outputs
  no_flow_control : if g_use_src_in_ready = false generate
    snk_out <= c_dp_siso_rdy;
    src_out <= r.src_out;
  end generate;

  gen_use_src_in_ready : if g_use_src_in_ready = true generate
    gen_dp_pipeline : if g_use_dp_latency_adapter = false generate
      u_dp_pipeline : entity work.dp_pipeline
      port map (
        rst          => rst,
        clk          => clk,
        -- ST sink
        snk_out      => snk_out,
        snk_in       => d.src_out,
        -- ST source
        src_in       => src_in,
        src_out      => src_out
      );
    end generate;

    gen_dp_latency_adapter : if g_use_dp_latency_adapter = true generate
      snk_out <= dp_latency_adapter_snk_out;
      dp_latency_adapter_snk_in <= r.src_out;

      u_dp_latency_adapter : entity work.dp_latency_adapter
      generic map (
        g_in_latency  => 2,
        g_out_latency => 1
      )
      port map (
        rst          => rst,
        clk          => clk,
        -- ST sink
        snk_out      => dp_latency_adapter_snk_out,
        snk_in       => dp_latency_adapter_snk_in,
        -- ST source
        src_in       => dp_latency_adapter_src_in,
        src_out      => dp_latency_adapter_src_out
      );

      dp_latency_adapter_src_in <= src_in;
      src_out <= dp_latency_adapter_src_out;
    end generate;
  end generate;

  -- p_reg
  r <= d when rising_edge(clk);

  p_comb : process(rst, r, snk_in)
    variable v : t_reg;
  begin
    -- Default
    v := r;

    -- Function
    --                   _                 _                 _
    --   snk_in.sop   __| |_______________| |_______________| |_______________|
    --                                  _                 _                 _
    --   snk_in.eop   _________________| |_______________| |_______________| |_
    --                   _  _  _  _  _  _  _  _  _  _  _  _  _  _  _  _  _  _
    --   snk_in.valid __|
    --   v.val_cnt        0  1  0  1  0  1  0  1  0  1  0  1  0  1  0  1  0  1
    --   v.pkt_cnt        0     1     2     0     1     2     0     1     2
    --                    ______________   _______________   _______________   _
    --   v.busy        __|              |_|               |_|               |_|
    --                    _     _     _     _     _     _     _     _     _
    --   v.src_out.sop __|0|___|1|___|2|___|0|___|1|___|2|___|0|___|1|___|2|___|
    --                       _     _     _     _     _     _     _     _     _
    --   v.src_out.eop _____|0|___|1|___|2|___|0|___|1|___|2|___|0|___|1|___|2|_

    -- input valid counter
    if snk_in.sop = '1' then
      v.val_cnt := 0;
    elsif snk_in.valid = '1' then
      if r.val_cnt < g_pkt_len - 1 then
        v.val_cnt := r.val_cnt + 1;
      else
        v.val_cnt := 0;
      end if;
    end if;

    -- output packet counter
    if snk_in.sop = '1' then
      v.pkt_cnt := 0;
    elsif snk_in.valid = '1' and v.val_cnt = 0 then
      v.pkt_cnt := r.pkt_cnt + 1;
    end if;

    -- output packet bsn, sync, sop and eop
    -- . Default passes on snk_in.sync, valid, data, re, im
    -- . The sync, valid can be passed on from snk_in, without checking
    --   src_in.ready, because the fact that snk_in.valid is '1' is
    --   sufficient for src_out.valid to be '1' too. The RL is taken care
    --   of by gen_use_src_in_ready.
    v.src_out         := snk_in;  -- passes on snk_in.sync, data, re, im
    v.src_out.bsn     := r.src_out.bsn;
    v.src_out.channel := r.src_out.channel;
    v.src_out.empty   := r.src_out.empty;
    v.src_out.err     := r.src_out.err;
    v.src_out.sop     := '0';
    v.src_out.eop     := '0';

    -- . output sop, eop
    if snk_in.valid = '1' then
      if v.val_cnt = 0 then
        v.src_out.sop := '1';
      end if;
      if v.val_cnt = g_pkt_len - 1 then
        v.src_out.eop := '1';
      end if;
    end if;

    -- . output bsn
    if snk_in.sop = '1' then
      v.src_out.bsn := snk_in.bsn;
    elsif v.src_out.sop = '1' then
      v.src_out.bsn := incr_uvec(r.src_out.bsn, g_bsn_increment);
    end if;

    -- . output channel
    v.src_out.channel := TO_DP_CHANNEL(v.pkt_cnt);

    -- . output err, empty. assume they are valid during entire input packet,
    --   so not only at the snk_in.eop
    if snk_in.valid = '1' then
      v.src_out.empty := snk_in.empty;
      v.src_out.err   := snk_in.err;
    end if;

    -- Synchronous reset
    if rst = '1' then
      v := c_reg_rst;
    end if;

    d <= v;
  end process;
end rtl;
