--------------------------------------------------------------------------------
--
-- Copyright (C) 2015
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Purpose:
--   The dp_repack_data works both as packer and as unpacker.
--
-- Block diagram:
--
-- A) Functional
--   The drawing shows g_in_nof_words=4 and g_out_nof_words=2 as example:
--
--                dp_repack_in               dp_repack_out
--                     ___                        ___
--                    |   |      pack_sosi       |   |--> src_out
--                    | 3 |--------------------->| 1 |
--                    |   |                      |   |
--                    | 2 | ^                    |   | ^
--                    |   | |valid               |   | |shift
--                    | 1 | |flush               | 0 | |
--                    |   | |                    |   | |
--                    | 0 |                      |   |
--   snk_in        -->|___|      pack_siso       |___|
--   snk_out.ready <--     <---------------------     <-- src_in.ready
--   snk_out.xon   <------------------------------------- src_in.xon
--
--
-- B) Flow control
--
--      RL=1                        RL=1               RL=1
--        .                           .                  .
--        .        /-----------------------------------------\
--        .        |                  .           _____  .   |
--        .        |   /------\     nxt_r        |     | r   |
--        .        \-->|      |---*-------*----->|p_reg|-----*---> src_out
--        .            |      |   |       |      |_____|
--     snk_in -------->|p_comb|<--|-------|--------------*-------- src_in
--                     |      |   |       |              |
--                     |      |   |       v              |
--                     |      |   |   /-------\          |
--                     |      |   |   |p_flow |          |
--                     \------/   |   \-------/          |
--                                |       |              |
--           nxt_r.hold_out.valid |       |              |
--                                v       |              |
--                                /|      |r_snk_out     |
--                               |0|------/              |
--    snk_out <------------------| |                     |
--                               |1|---------------------/
--                                \|
--
-- Description:
--   The dp_repack_data repacks g_in_nof_words of width g_in_dat_w into
--   g_out_nof_words of width g_out_dat_w.
--
-- . g_in_bypass, g_out_bypass
--   The dp_repack_in and dp_repack_out can be bypassed to save logic and to
--   avoid the pipeline stage. Default both are FALSE, but they can be set
--   TRUE if:
--
--   . g_in_bypass =TRUE if g_in_nof_words=g_out_nof_words or g_in_nof_words=1
--   . g_out_bypass=TRUE if g_in_nof_words=g_out_nof_words or g_out_nof_words=1
--
--   Both the dp_repack_in and dp_repack_out stage do work correctly independent
--   of the g_*_bypass setting. When g_*_bypass=FALSE then they merely
--   add a transparant pipeline delay. It is important that they also work for
--   g_*_bypass=FALSE because that gives confidence that their implementation
--   structure is ok.
--
-- . g_in_nof_words and input block size
--   The input block size in words is indicated by snk_in.sop and snk_in.eop.
--   Each subsection of g_in_nof_words is packed into g_out_nof_words. The
--   input block size does not have to be a multiple of g_in_nof_words. When
--   the snk_in.eop occurs the last repack is initiated without need for input
--   data padding. If the block length is an integer multiple of
--   g_in_nof_words then the dp_repack_data introduces no gaps between blocks.
--   If the block length is a fractional multiple of g_in_nof_words then there
--   will be a gap after the block due to that the dp_repack_in needs to
--   shift up the last subsection for the 'missing' input words.
--
-- . g_in_dat_w*g_in_nof_words <, =, > g_in_dat_w*g_in_nof_words
--   . = : no subsection zero padding
--   . < : the subsections will be zero padded
--   . > : then the input must have sufficient zero padded bits per
--         subsection that can be stripped without data loss.
--
-- . Resolution of the empty field
--   The g_in_symbol_w is used to define the resolution of snk_in.empty and
--   the g_out_symbol_w is used to define the resolution of src_out.empty. If
--   they are 1 then the resolution is in number of bits, because the symbol
--   size is then 1 bit. Their value has no effect on the packed data it self,
--   only on the meaning of the empty field. Hence if the empty field is not
--   used, then the setting of g_in_symbol_w and g_out_symbol_w is dont care.
--
-- Remarks:
-- . Originally reused from LOFAR rad_repack.vhd and rad_repack(rtl).vhd. This
--   dp_repack_data still uses the shift in input register in and the shift out
--   output register, but no longer the intermediate buffer register.
--   Using shift in and shift out may ease timing closure because the routing
--   is more local compared to using a demultiplexer to put the input data in
--   the input register and a multiplexer to get the data directly from the
--   output register. For the demultiplexer / multiplexer it would be possible
--   to only use one internal register.
--   Using shift up is sufficient, the shift down option is not needed. With
--   shift up the data is input a [0] and output the high index.
--   Instead of requiring an snk_in.valid duty cycle this dp_repack_data uses
--   snk_out.ready flow control and can handle src_in.ready flow control.
--
-- . To pack ETH/IP/UDP header slv of 14 + 20 + 8 = 42 octets into 32 bit words
--   use:
--     u_dp_repack_data : ENTITY dp_lib.dp_repack_data
--     GENERIC MAP (
--       g_in_bypass         => TRUE,
--       g_in_dat_w          => 8 * 42,
--       g_in_nof_words      => 1,
--       g_in_symbol_w       => 8,
--       g_out_bypass        => FALSE,
--       g_out_dat_w         => 32,
--       g_out_nof_words     => 11,
--       g_out_symbol_w      => 8
--     )
--   The src_out.empty will be 2, because:
--     (g_out_dat_w*g_out_nof_words-g_in_dat_w*g_in_nof_words)/g_out_symbol_w
--      = (32*11 - 42*8*1)/ 8 = 2 octet symbols
-- . Instead of using dp_sosi.data for intermediate results in dp_repack_data,
--   a seperate std_logic_vector is used to carry the sosi data as the required
--   vector width can become larger than c_dp_stream_data_w. Note that for the
--   in/out sosi of dp_repack_data, the dp_sosi.data field is still used such
--   that there is no added complexity for the user of dp_repack_data.
--
-- Design steps:
-- * In total the development took 5 days. On day 3 I was in distress because
--   I could not get it to work so I needed to rethink. After changing to the
--   new flow control scheme that uses nxt_r the design was gradually improved
--   by getting the dp_repack_data instances in tb_tb_dp_repack_data to work one
--   by one. First only for e_active stimuli and later also for e_random and
--   e_pulse. Initially about 80 % of the functionality was implemented but
--   and subsequently each feature was verified starting with the basic
--   features and then themore detailed features. This step by step approach
--   makes that the bugs appear one by one instead of all together. Without a
--   step by step approach the bugs are too big to solve.
--   . First cases with g_in_nof_words=1 and g_out_nof_words were made to work
--     for different g_pkt_len <, =, > g_in_nof_words.
--   . Then the empty functionality for g_pkt_len MOD g_in_nof_words /= 0 was
--     added.
--   . Tried g_out_dat_w=1 which makes dp_repack_data a serializer/deserializer.
--   . Then apply external flow control using c_dp_flow_control_enum_arr in
--     the tb_tb_dp_repack_data was verified resulting in small corrections.
--   . Then verified g_in_dat_w * g_in_nof_words > or < g_out_dat_w *
--     g_out_nof_words which require padding in the subsection. The > case
--     occurs for packing and the < case then occurs for unpacking.
--   . Added g_bypass to force using wires instead of a void dp_repack_in or
--     dp_repack_out stage.
--   . Verified g_in_symbol_w and g_out_symbol_w /= 1.
-- * The development used the tb_dp_repack_data testbench that does a pack and
--   an unpack to be able to verify the data. The c_no_unpack and
--   c_enable_repack_in and c_enable_repack_out parameters in the tb are
--   useful to be able to isolate a component for debugging.

library IEEE, common_lib, dp_lib;
use IEEE.std_logic_1164.all;
use common_lib.common_pkg.all;
use dp_lib.dp_stream_pkg.all;

entity dp_repack_in is
  generic (
    g_bypass          : boolean := false;
    g_in_dat_w        : natural;
    g_in_nof_words    : natural;
    g_in_symbol_w     : natural := 1  -- default 1 for snk_in.empty in nof bits, else use power of 2
  );
  port (
    rst              : in  std_logic;
    clk              : in  std_logic;

    snk_out          : out t_dp_siso;
    snk_in           : in  t_dp_sosi;

    src_in           : in  t_dp_siso;
    src_out          : out t_dp_sosi;
    src_out_data     : out std_logic_vector(g_in_dat_w * g_in_nof_words - 1 downto 0)
  );
end dp_repack_in;

architecture rtl of dp_repack_in is
  constant c_in_buf_dat_w      : natural := g_in_dat_w * g_in_nof_words;
  constant c_bit_cnt_max       : natural := c_in_buf_dat_w;
  constant c_in_empty_lo       : natural := true_log2(g_in_symbol_w);

  type t_dat_arr  is array (integer range <>) of std_logic_vector(g_in_dat_w - 1 downto 0);

  type t_reg is record
    dat_arr       : t_dat_arr(g_in_nof_words - 1 downto 0);  -- internally use dat_arr[] to represent v.src_out.data
    src_out       : t_dp_sosi;  -- sosi output
    src_out_data  : std_logic_vector(c_in_buf_dat_w - 1 downto 0);  -- Use seperate STD_LOGIC_VECTOR to carry the sosi data as c_in_buf_dat_w can be larger than c_dp_stream_data_w.
    hold_out      : t_dp_sosi;  -- hold snk_in.sync/sop/eop until end of section and then hold valid src_out until src_in.ready
    flush         : std_logic;  -- shift when snk_in.valid or flush in case the last subsection has < g_in_nof_words
    dat_bit_cnt   : natural range 0 to c_bit_cnt_max;  -- actual nof bits in subsection
    pack_bit_cnt  : natural range 0 to c_bit_cnt_max;  -- count nof bits in subsection
  end record;

  signal data_vec   : std_logic_vector(c_in_buf_dat_w - 1 downto 0);

  signal r_snk_out  : t_dp_siso := c_dp_siso_rdy;
  signal r          : t_reg;
  signal nxt_r      : t_reg;

  -- Debug signals
  signal snk_in_data        : std_logic_vector(g_in_dat_w - 1 downto 0);
  signal i_src_out          : t_dp_sosi;
  signal i_src_out_data     : std_logic_vector(c_in_buf_dat_w - 1 downto 0);

  signal dbg_g_in_dat_w   : natural := g_in_dat_w;
  signal dbg_in_nof_words : natural := g_in_nof_words;
  signal dbg_in_symbol_w  : natural := g_in_symbol_w;
  signal dbc_in_buf_dat_w : natural := c_in_buf_dat_w;
begin
  snk_in_data <= snk_in.data(g_in_dat_w - 1 downto 0);

  src_out      <= i_src_out;
  src_out_data <= i_src_out_data;

  gen_bypass : if g_bypass = true generate
    snk_out        <= src_in;
    i_src_out      <= snk_in;
    i_src_out_data <= RESIZE_UVEC(snk_in.data, c_in_buf_dat_w);
  end generate;

  no_bypass : if g_bypass = false generate
    p_comb : process(rst, r, snk_in, data_vec, src_in)
      variable v : t_reg;
    begin
      ------------------------------------------------------------------------
      -- Default
      v := r;
      v.src_out.sync  := '0';
      v.src_out.valid := '0';
      v.src_out.sop   := '0';
      v.src_out.eop   := '0';

      --------------------------------------------------------------------------
      -- Function
      if r.hold_out.valid = '0' then

        -- Clear hold_out for new output valid (= new subsection)
        if r.src_out.valid = '1' then
          v.hold_out := c_dp_sosi_rst;
        end if;

        -- Capture the snk_in block info that is valid at sop and eop
        if snk_in.sop = '1' then
          v.hold_out.sop    := '1';
          v.hold_out.sync   := snk_in.sync;
          v.src_out.bsn     := snk_in.bsn;
          v.src_out.channel := snk_in.channel;
        end if;
        if snk_in.eop = '1' then
          v.hold_out.eop    := '1';
          v.hold_out.empty  := SHIFT_UVEC(snk_in.empty, -c_in_empty_lo);  -- use snk_in.empty as offset for src_out.empty in nof bits
          v.src_out.err     := snk_in.err;
        end if;

        -- Capture the data per subsection in a block
        if snk_in.valid = '1' or r.flush = '1' then
          -- shift in during block
          v.dat_arr(g_in_nof_words - 1 downto 1) := r.dat_arr(g_in_nof_words - 2 downto 0);  -- shift up from low to high and shift in at index 0
          if r.flush = '1' then
            v.dat_arr(0) := (others => '0');  -- shift in data=0 for flush
          else
            v.dat_arr(0) := snk_in.data(g_in_dat_w - 1 downto 0);  -- shift in valid data
          end if;

          -- pack subsection
          if r.pack_bit_cnt < c_in_buf_dat_w - g_in_dat_w then
            v.pack_bit_cnt := r.pack_bit_cnt + g_in_dat_w;

            -- early end of pack subsection
            if snk_in.eop = '1' then
              v.flush := '1';  -- enable flush in case eop occurs before end of pack subsection
              v.dat_bit_cnt := v.pack_bit_cnt;  -- capture the current subsection pack_bit_cnt
            end if;
          else  -- r.pack_bit_cnt=c_in_buf_dat_w-g_in_dat_w
            -- default end of pack subsection
            v.pack_bit_cnt := 0;
            v.flush := '0';
            if r.flush = '0' then
              v.dat_bit_cnt := c_in_buf_dat_w;  -- set default subsection pack_bit_cnt
            end if;

            v.hold_out.valid := '1';  -- the function has new data to output
          end if;
        end if;

        -- pass on the v.dat_arr as data vector
        v.src_out_data := data_vec;
        v.src_out.data := RESIZE_DP_DATA(data_vec);

        -- pass on dat_bit_cnt via DP empty field
        v.src_out.empty := INCR_UVEC(v.hold_out.empty, c_in_buf_dat_w - v.dat_bit_cnt);

        -- output input stage into output stage when ready, else hold_out.valid to signal pending output
        if v.hold_out.valid = '1' then
          if src_in.ready = '1' then
            v.src_out.valid := '1';
            v.src_out.sync  := v.hold_out.sync;
            v.src_out.sop   := v.hold_out.sop;
            v.src_out.eop   := v.hold_out.eop;
            v.hold_out.valid := '0';
          end if;
        end if;
      else
        -- pending output
        if src_in.ready = '1' then
          v.src_out.valid := '1';
          v.src_out.sync  := r.hold_out.sync;
          v.src_out.sop   := r.hold_out.sop;
          v.src_out.eop   := r.hold_out.eop;
          v.hold_out.valid := '0';
        end if;
      end if;

      ------------------------------------------------------------------------
      -- Reset and nxt_r
      if rst = '1' then
        v.src_out       := c_dp_sosi_rst;
        v.hold_out      := c_dp_sosi_rst;
        v.flush         := '0';
        v.dat_bit_cnt   := 0;
        v.pack_bit_cnt  := 0;
      end if;

      nxt_r <= v;
    end process;

    --------------------------------------------------------------------------
    -- p_reg
    r <= nxt_r when rising_edge(clk);

    --------------------------------------------------------------------------
    -- Wires
    p_data_vec : process(nxt_r)
    begin
      for I in 0 to g_in_nof_words - 1 loop
        data_vec((I + 1) * g_in_dat_w - 1 downto I * g_in_dat_w) <= nxt_r.dat_arr(I);
      end loop;
    end process;

    --------------------------------------------------------------------------
    -- Wired output
    i_src_out <= r.src_out;
    i_src_out_data <= r.src_out_data;
    --------------------------------------------------------------------------
    -- Flow control

    -- local function flow control
    p_flow : process(nxt_r)
    begin
      r_snk_out <= c_dp_siso_rdy;
      if nxt_r.flush = '1' then
        r_snk_out.ready <= '0';  -- input shift in stage function is always ready except when flushing
      end if;
    end process;

    -- combined local and remote src_in flow control
    snk_out.ready <= r_snk_out.ready when nxt_r.hold_out.valid = '0' else src_in.ready;  -- if there is pending output then the src_in ready determines the flow control
    snk_out.xon   <= src_in.xon;  -- just pass on the xon/off frame flow control
  end generate;

end rtl;

library IEEE, common_lib, dp_lib;
use IEEE.std_logic_1164.all;
use common_lib.common_pkg.all;
use dp_lib.dp_stream_pkg.all;

entity dp_repack_out is
  generic (
    g_bypass           : boolean := false;
    g_in_buf_dat_w     : natural;
    g_out_dat_w        : natural;
    g_out_nof_words    : natural;
    g_out_symbol_w     : natural := 1  -- default 1 for snk_in.empty in nof bits, else use power of 2
  );
  port (
    rst              : in  std_logic;
    clk              : in  std_logic;

    snk_out          : out t_dp_siso;
    snk_in           : in  t_dp_sosi;
    snk_in_data      : in  std_logic_vector(g_in_buf_dat_w - 1 downto 0);

    src_in           : in  t_dp_siso;
    src_out          : out t_dp_sosi
  );
end dp_repack_out;

architecture rtl of dp_repack_out is
  constant c_out_buf_dat_w     : natural := g_out_dat_w * g_out_nof_words;
  constant c_out_buf_dat_lo    : natural := sel_a_b(c_out_buf_dat_w > g_in_buf_dat_w, c_out_buf_dat_w -  g_in_buf_dat_w, 0);  -- pack into subsection with 0 or more padding bits
  constant c_snk_in_dat_lo     : natural := sel_a_b(c_out_buf_dat_w < g_in_buf_dat_w,  g_in_buf_dat_w - c_out_buf_dat_w, 0);  -- unpack from subsection that has 0 or more padding bits
  constant c_bit_cnt_max       : natural := c_out_buf_dat_w;
  constant c_out_empty_lo      : natural := true_log2(g_out_symbol_w);

  type t_dat_arr is array (integer range <>) of std_logic_vector(g_out_dat_w - 1 downto 0);

  type t_reg is record
    dat_arr       : t_dat_arr(g_out_nof_words - 1 downto 0);
    src_out       : t_dp_sosi;
    hold_out      : t_dp_sosi;  -- hold src_out valid and sync/sop/eop until src_in.ready
    shift         : std_logic;  -- shift out the dat_arr
    dat_bit_cnt   : natural range 0 to c_bit_cnt_max;  -- actual nof bits in subsection
    pack_bit_cnt  : natural range 0 to c_bit_cnt_max;  -- count nof bits in subsection
    empty_bit_cnt : natural range 0 to c_bit_cnt_max;  -- empty nof bits in subsection
    eos           : std_logic;  -- end of subsection
  end record;

  signal data_vec  : std_logic_vector(c_out_buf_dat_w - 1 downto 0) := (others => '0');

  signal r_snk_out : t_dp_siso := c_dp_siso_rdy;
  signal r         : t_reg;
  signal nxt_r     : t_reg;

  -- Debug signals
  signal i_src_out          : t_dp_sosi;
  signal src_out_data       : std_logic_vector(g_out_dat_w - 1 downto 0);

  signal dbg_g_in_buf_dat_w : natural := g_in_buf_dat_w;
  signal dbg_g_out_dat_w    : natural := g_out_dat_w;
  signal dbg_out_nof_words  : natural := g_out_nof_words;
  signal dbg_out_symbol_w   : natural := g_out_symbol_w;
  signal dbc_out_buf_dat_w  : natural := c_out_buf_dat_w;
  signal dbc_out_buf_dat_lo : natural := c_out_buf_dat_lo;
  signal dbc_snk_in_dat_lo  : natural := c_snk_in_dat_lo;
begin
  src_out      <= i_src_out;
  src_out_data <= i_src_out.data(g_out_dat_w - 1 downto 0);

  gen_bypass : if g_bypass = true generate
    snk_out <= src_in;

    p_src_out : process(snk_in)
    begin
      i_src_out <= snk_in;
      if c_snk_in_dat_lo > 0 then
        i_src_out.data  <= SHIFT_UVEC(RESIZE_DP_DATA(snk_in_data), c_snk_in_dat_lo);
        i_src_out.empty <= INCR_UVEC( snk_in.empty, -c_snk_in_dat_lo);
      end if;
      if c_out_buf_dat_lo > 0 then
        i_src_out.data  <= SHIFT_UVEC(RESIZE_DP_DATA(snk_in_data), -c_out_buf_dat_lo);
        i_src_out.empty <= INCR_UVEC( snk_in.empty, c_out_buf_dat_lo);
      end if;
    end process;
  end generate;

  no_bypass : if g_bypass = false generate
    p_comb : process(rst, snk_in, r, data_vec, src_in)
      variable v : t_reg;
    begin
      ------------------------------------------------------------------------
      -- Default
      v := r;
      v.src_out.sync  := '0';
      v.src_out.valid := '0';
      v.src_out.sop   := '0';
      v.src_out.eop   := '0';

      ------------------------------------------------------------------------
      -- Function
      if r.hold_out.valid = '0' then

        -- Clear hold_out for new output valid
        if r.src_out.sop = '1' then
          v.hold_out.sync := '0';
          v.hold_out.sop  := '0';
        end if;
        if r.src_out.eop = '1' then
          v.hold_out.eop := '0';
        end if;

        -- Capture the snk_in block info that is valid at sop and eop
        if snk_in.sop = '1' then
          v.hold_out.sop    := '1';
          v.hold_out.sync   := snk_in.sync;
          v.src_out.bsn     := snk_in.bsn;
          v.src_out.channel := snk_in.channel;
        end if;
        if snk_in.eop = '1' then
          v.hold_out.eop    := '1';  -- local function will calculate src_out.empty based on snk_in.empty
          v.src_out.err     := snk_in.err;
        end if;

        if r.shift = '1' then
          -- shift out rest of subsection
          v.hold_out.valid := '1';

          v.dat_arr(g_out_nof_words - 1 downto 1) := r.dat_arr(g_out_nof_words - 2 downto 0);  -- shift up from low to high and shift out at high index
          v.dat_arr(0) := (others => '0');  -- shift in data=0

          v.pack_bit_cnt := r.pack_bit_cnt - g_out_dat_w;

          -- end of pack subsection
          if v.pack_bit_cnt <= r.empty_bit_cnt then
            v.eos   := '1';  -- end of subsection, so ready for new snk_in
            v.shift := '0';  -- stop shifting
          end if;

        elsif snk_in.valid = '1' then
          -- start of pack subsection
          v.hold_out.valid := '1';

          for I in 0 to g_out_nof_words - 1 loop
            v.dat_arr(I) := data_vec((I + 1) * g_out_dat_w - 1 downto I * g_out_dat_w);
          end loop;

          v.dat_bit_cnt := g_in_buf_dat_w - c_snk_in_dat_lo;  -- default dat_bit_cnt per subsection
          if snk_in.eop = '1' then
            v.dat_bit_cnt := g_in_buf_dat_w - TO_UINT(snk_in.empty);  -- pass on last subsection dat_bit_cnt info via DP empty field
          end if;

          v.pack_bit_cnt  := c_out_buf_dat_w - g_out_dat_w;
          v.empty_bit_cnt := c_out_buf_dat_w - v.dat_bit_cnt;
          v.eos           := '0';
          v.shift         := '1';

          -- end of pack subsection
          if v.pack_bit_cnt <= v.empty_bit_cnt then
            v.eos   := '1';  -- end of subsection, so ready for new snk_in
            v.shift := '0';
          end if;
        end if;

        -- fill in local empty if this is the last subsection of a block
        if v.eos = '1' then
          if v.hold_out.eop = '1' then
            v.src_out.empty := TO_DP_EMPTY(v.empty_bit_cnt - v.pack_bit_cnt);  -- in nof bits
            v.src_out.empty := SHIFT_UVEC(v.src_out.empty, c_out_empty_lo);  -- in nof symbols
          end if;
        end if;

        -- pass on the v.dat_arr as data vector
        v.src_out.data  := RESIZE_DP_DATA(v.dat_arr(g_out_nof_words - 1));

        -- output valid data when ready, else hold_out.valid to signal pending output
        if v.hold_out.valid = '1' then
          if src_in.ready = '1' then
            v.src_out.valid  := '1';
            v.src_out.sync   := v.hold_out.sync;
            v.src_out.sop    := v.hold_out.sop;
            v.src_out.eop    := v.hold_out.eop and v.eos;  -- output eop at end of subsection
            v.hold_out.valid := '0';
          end if;
        end if;

      else
        -- pending output
        if src_in.ready = '1' then
          v.src_out.valid := '1';
          v.src_out.sync  := r.hold_out.sync;
          v.src_out.sop   := r.hold_out.sop;
          v.src_out.eop   := r.hold_out.eop and r.eos;  -- output eop at end of subsection
          v.hold_out.valid := '0';
        end if;
      end if;

      ------------------------------------------------------------------------
      -- Reset and nxt_r
      if rst = '1' then
        v.src_out       := c_dp_sosi_rst;
        v.hold_out      := c_dp_sosi_rst;
        v.shift         := '0';
        v.dat_bit_cnt   := 0;
        v.pack_bit_cnt  := 0;
        v.empty_bit_cnt := 0;
        v.eos           := '0';
      end if;

      nxt_r <= v;
    end process;

    --------------------------------------------------------------------------
    -- p_reg
    r <= nxt_r when rising_edge(clk);

    --------------------------------------------------------------------------
    -- Wires
    data_vec(c_out_buf_dat_w - 1 downto c_out_buf_dat_lo) <= snk_in_data(g_in_buf_dat_w - 1 downto c_snk_in_dat_lo);

    --------------------------------------------------------------------------
    -- Wired output
    i_src_out <= r.src_out;

    --------------------------------------------------------------------------
    -- Flow control

    -- local function flow control
    p_flow : process(nxt_r)
    begin
      r_snk_out <= c_dp_siso_rdy;
      if nxt_r.shift = '1' and nxt_r.eos = '0' then
        r_snk_out.ready <= '0';  -- output shift out stage function is only ready when it is not shifting or at the end of the subsection
      end if;
    end process;

    -- combined local and remote src_in flow control
    snk_out.ready <= r_snk_out.ready when nxt_r.hold_out.valid = '0' else src_in.ready;  -- if there is pending output then the src_in ready determines the flow control
    snk_out.xon   <= src_in.xon;  -- just pass on the xon/off frame flow control
  end generate;

end rtl;

library IEEE, common_lib, dp_lib;
use IEEE.std_logic_1164.all;
use common_lib.common_pkg.all;
use dp_lib.dp_stream_pkg.all;

entity dp_repack_data is
  generic (
    g_enable_repack_in  : boolean := true;
    g_enable_repack_out : boolean := true;
    g_in_bypass         : boolean := false;
    g_in_dat_w          : natural;
    g_in_nof_words      : natural;
    g_in_symbol_w       : natural := 1;  -- default 1 for snk_in.empty in nof bits, else use power of 2
    g_out_bypass        : boolean := false;
    g_out_dat_w         : natural;
    g_out_nof_words     : natural;
    g_out_symbol_w      : natural := 1;  -- default 1 for src_out.empty in nof bits, else use power of 2
    g_pipeline_ready    : boolean := false  -- TRUE to pipeline ready from src_in to snk_out.
  );
  port (
    rst              : in  std_logic;
    clk              : in  std_logic;

    snk_out          : out t_dp_siso;
    snk_in           : in  t_dp_sosi;

    src_in           : in  t_dp_siso := c_dp_siso_rdy;
    src_out          : out t_dp_sosi
  );
end dp_repack_data;

architecture str of dp_repack_data is
  constant c_in_buf_dat_w      : natural := g_in_dat_w * g_in_nof_words;

  signal snk_in_data       : std_logic_vector(g_in_dat_w - 1 downto 0);
  signal i_snk_out         : t_dp_siso;
  signal i_snk_in          : t_dp_sosi;

  signal pack_siso         : t_dp_siso;
  signal pack_sosi         : t_dp_sosi;
  signal pack_sosi_data    : std_logic_vector(c_in_buf_dat_w - 1 downto 0);

  signal i_src_out         : t_dp_sosi;
  signal src_out_data      : std_logic_vector(g_out_dat_w - 1 downto 0);

  signal snk_out_ready_reg : std_logic_vector(0 to c_dp_stream_rl);
  signal pack_ready_reg    : std_logic_vector(0 to c_dp_stream_rl);
begin
  gen_dp_pipeline_ready: if g_pipeline_ready = true generate
    u_dp_pipeline_ready : entity dp_lib.dp_pipeline_ready
    port map (
      rst     => rst,
      clk     => clk,

      snk_out => snk_out,
      snk_in  => snk_in,

      src_in  => i_snk_out,
      src_out => i_snk_in
    );
  end generate;

  gen_no_dp_pipeline_ready: if g_pipeline_ready = false generate
    i_snk_in <= snk_in;
    snk_out  <= i_snk_out;
  end generate;

  src_out <= i_src_out;

  snk_in_data    <= i_snk_in.data(g_in_dat_w - 1 downto 0);
  src_out_data   <= i_src_out.data(g_out_dat_w - 1 downto 0);

  no_dp_repack_in : if g_enable_repack_in = false generate
    i_snk_out <= pack_siso;
    pack_sosi <= i_snk_in;
    pack_sosi_data <= RESIZE_UVEC(snk_in_data, c_in_buf_dat_w);
  end generate;

  gen_dp_repack_in : if g_enable_repack_in = true generate
    u_dp_repack_in : entity work.dp_repack_in
    generic map (
      g_bypass       => g_in_bypass,
      g_in_dat_w     => g_in_dat_w,
      g_in_nof_words => g_in_nof_words,
      g_in_symbol_w  => g_in_symbol_w
    )
    port map (
      rst      => rst,
      clk      => clk,

      snk_out  => i_snk_out,
      snk_in   => i_snk_in,

      src_in   => pack_siso,
      src_out  => pack_sosi,
      src_out_data => pack_sosi_data
    );
  end generate;

  no_dp_repack_out : if g_enable_repack_out = false generate
    pack_siso <= src_in;
    i_src_out <= pack_sosi;
  end generate;

  gen_dp_repack_out : if g_enable_repack_out = true generate
    u_dp_repack_out : entity work.dp_repack_out
    generic map (
      g_bypass        => g_out_bypass,
      g_in_buf_dat_w  => c_in_buf_dat_w,
      g_out_dat_w     => g_out_dat_w,
      g_out_nof_words => g_out_nof_words,
      g_out_symbol_w  => g_out_symbol_w
    )
    port map (
      rst      => rst,
      clk      => clk,

      snk_out  => pack_siso,
      snk_in   => pack_sosi,
      snk_in_data => pack_sosi_data,

      src_in   => src_in,
      src_out  => i_src_out
    );
  end generate;

  -- Simulation only: internal stream RL verification
  proc_dp_siso_alert(clk, i_snk_in, i_snk_out, snk_out_ready_reg);
  proc_dp_siso_alert(clk, pack_sosi, pack_siso, pack_ready_reg);
end str;
